#include <bitset>
#include <cassert> 

#include <cfloat> // 'DBL_MAX'...
#include <climits> // 'CHAR_BIT', SHRT_MAX'...
#include <cmath> // 'isinf', 'pow'...
#include <iostream>
#include <type_traits>

#include "Data_structure.h"

int main() {
	My_data_structure object;

	std::cout << "bool: " << sizeof(bool) << '\n';
	std::cout << "char: " << sizeof(object.char_attribute) << '\n';
	std::cout << "short: " << sizeof(short) << '\n';
	std::cout << "int: " << sizeof(int) << '\n';
	std::cout << "long: " << sizeof(long) << '\n';
	std::cout << "long long: " << sizeof(long long) << '\n';

	std::cout << "float: " << sizeof(float) << '\n';
	std::cout << "double: " << sizeof(double) << '\n';
	std::cout << "long double: " << sizeof(long double) << '\n';
	std::cout << "My_data_structure: " << sizeof(My_data_structure) << "\n\n";

	object.bool_attribute = true;
	std::cout << object.bool_attribute << '\n';
	object.char_attribute = '\r';
	std::cout << object.char_attribute << '\n';
	object.unsigned_char_attribute = 'a';
	std::cout << object.unsigned_char_attribute << '\n';

	object.int_attribute = (short)0xFFFF;
	std::cout << object.int_attribute << '\n';
	object.unsigned_short_attribute = 0xFFFF;
	std::cout << object.unsigned_short_attribute << '\n';
	object.int_attribute = (int)6.5E1;
	std::cout << object.int_attribute << '\n';
	object.unsigned_int_attribute = (unsigned int)7.2e-2;
	std::cout << object.unsigned_int_attribute << '\n';
	object.long_attribute = 500000L;
	std::cout << object.long_attribute << '\n';
	object.unsigned_long_attribute = 0xFFFFFFFE + 1L;
	std::cout << object.unsigned_long_attribute << '\n';
	object.long_long_attribute = 9LL;
	std::cout << object.long_long_attribute << '\n';
	object.unsigned_long_long_attribute = 0xFFFFFFFFFFFFFFFE + 1LL;
	std::cout << object.unsigned_long_long_attribute << '\n';

	object.float_attribute = 1.F;
	std::cout << object.float_attribute << '\n';
	object.double_attribute = 1.;
	std::cout << object.double_attribute << '\n';
	object.long_double_attribute = 1.L;
	std::cout << object.long_double_attribute << '\n';

	object.union_attribute.is_char = 'A';
	std::cout << "'A' as 'char': " << object.union_attribute.is_char << '\n';
	std::cout << "'A' as 'int': " << object.union_attribute.is_int << '\n';
	object.union_attribute.is_int = 66;
	std::cout << "'66' as 'char': " << object.union_attribute.is_char << '\n';
	std::cout << "'66' as 'int': " << object.union_attribute.is_int << '\n';

	std::bitset<CHAR_BIT> a{ 97 }; // 'a'
	std::byte a_{ 97 }; // 'a'

	// C++17:
	std::cout << std::boolalpha << "'std::is_aggregate_v<My_data_structure>': " << std::is_aggregate_v<My_data_structure> << std::endl; // 'true'
	// C++11 (included: bool, ..., long long) (excluded: enum, float...) :
	std::cout << std::boolalpha << "'std::is_integral<decltype(object.char_attribute)>': " << std::is_integral_v<decltype(object.char_attribute)> << std::endl; // 'true'

	// C++14:
	My_data_structure another_object{ false, 'a', 'a', 1, 1, 1, 1, 1L, 1L, 1LL, 1LL, 1.F, 1., 1.L,
		{1., 1.} };
	// C++20:
	My_data_structure yet_another_object{ .bool_attribute = false /*,.char_attribute = 'a'*/, .unsigned_char_attribute = 'b' };

	std::string FB[] = { "Franck", /* list initialization -> */
		{'B', 'a', 'r', 'b', 'i', 'e', 'r'} };
	wchar_t F[] = L"Franck";

	// https://docs.microsoft.com/en-us/cpp/cpp/left-shift-and-right-shift-operators-input-and-output?view=vs-2017
	std::bitset<sizeof(short) * CHAR_BIT > one(1);
	std::cout << '\n' << "'one': " << one << '\n'; // '0000 0000 0000 0001'
	// Bit positions that have been vacated by a shift operation are zero-filled:
	std::bitset<sizeof(short) * CHAR_BIT > two(1 << 1 /* 1-bit left shift */);
	std::cout << "'two': " << two << '\n'; // '0000 0000 0000 0010'
	std::bitset<sizeof(short) * CHAR_BIT > zero(1 >> 1 /* 1-bit right shift */);
	std::cout << "'zero': " << zero << '\n'; // '0000 0000 0000 0000'

	std::bitset<sizeof(short) * CHAR_BIT > minus_one(-1);
	std::cout << "'minus_one': " << minus_one << '\n'; // '1111 1111 1111 1111'

	// How negative numbers are represented (www.devx.com/tips/Tip/16134):
	std::bitset<sizeof(short) * CHAR_BIT > minus_minus_one(~minus_one); // ~'1111 1111 1111 1111' -> '0000 0000 0000 0000'
	minus_minus_one[0] = true; // '0000 0000 0000 0001'
	std::cout << "'minus_minus_one': " << minus_minus_one << '\n'; // '0000 0000 0000 0001'

	// If you left-shift a signed number so that the sign bit is affected, the result is undefined:
	std::bitset<sizeof(short) * CHAR_BIT > unknown(16384 << 1); // '0100 0000 0000 0000' << 1
	std::cout << "'unknown': " << unknown << '\n'; // '1000 0000 0000 0000'
	std::bitset<sizeof(short) * CHAR_BIT > short_min(SHRT_MIN); // '-32768'
	std::cout << "'short_min': " << short_min << '\n'; // '1000 0000 0000 0000'

	std::cout << '\n' << "'SHRT_MAX': " << SHRT_MAX << '\n' // '0111 1111 1111 1111' / 32767
		<< "'SHRT_MIN': " << SHRT_MIN << '\n'; // '1000 0000 0000 0000' / -32768
	assert(SHRT_MAX != -SHRT_MIN);
	assert(SHRT_MAX == ~SHRT_MIN);

	// This line compiles with g++, but it doesn't with Visual Studio:
	// std::cout << std::boolalpha << "std::isinf(1 / 0): " << std::isinf(1 / 0) << '\n'; // 'false' because division by zero is undefined here...

	// Some lines don't compile with Visual Studio ('div or mode by zero'):
	std::cout << '\n' << std::boolalpha
		<< "std::isnan(NAN): " << std::isnan(NAN) << '\n' // 'true'
		<< "std::isnan(INFINITY): " << std::isnan(INFINITY) << '\n' // 'false'
		<< "std::isnan(1 / 0.F): " << std::isnan(1 / 0.F) << '\n' // 'false'
		<< "std::isinf(1 / 0.F): " << std::isinf(1 / 0.F) << '\n' // 'true'
		<< "std::isnan(DBL_MAX * 2.): " << std::isnan(DBL_MAX * 2.) << '\n' // 'false'
		<< "std::isinf(DBL_MAX * 2.): " << std::isinf(DBL_MAX * 2.) << '\n' // 'true'
		<< "std::isnan(0 / 0.): " << std::isnan(0 / 0.) << '\n' // 'true'
		<< "std::isnan(INFINITY - INFINITY): " << std::isnan(INFINITY - INFINITY) << '\n'; // 'true' 

		/**
		 * Floating point numbers (www.learncpp.com/cpp-tutorial/25-floating-point-numbers/)
		 * IEEE 754 standard (www3.ntu.edu.sg/home/ehchua/programming/java/datarepresentation.html)
		 */
	union {
		float f = 1.1F; // '0 - 0111 1111 - 000 1100 1100 1100 1100 1101' -> 1 sign bit + 8 exponent bits + 23 fraction (significant) bits = 32 bits
		uint32_t i; // if 'sizeof (float) == 8' then 'uint64_t'
	} _one_point_one;
	// S == '0' -> positive
	// E == '0111 1111' -> 127
	// Fraction ==  '000 1100 1100 1100 1100 1101' -> in IEEE 754 standard, the actual fraction is normalized with an implicit leading '1', so
	// the actual fraction is 1.000 1100 1100 1100 1100 1101 = 1 + 2^-4 + 2^-5 + 2^-8 + 2^-9 + 2^-12 + 2^-13 +
	// 2^-16 + 2^-17 + 2^-20 + 2^-21 + 2^-23 = 1 + 1/16 + 1/32 + ... = 1.1

	// About E: in IEEE 754 standard, the actual exponent is 'E - 127' (so-called excess-127 or bias-127). This is because we need to represent both positive and negative exponent.
	// With an 8-bit E, ranging from 0 to 255, the excess-127 scheme could provide actual exponent of -127 to 128.

	/**
	 * Final result:
	 * 1.1 * 2^(E - 127) = 1.1 * 2^0 = 1.1
	 */

	float check = 1 + pow(2.F, -4.F) + pow(2.F, -5.F)
		+ pow(2.F, -8.F) + pow(2.F, -9.F)
		+ pow(2.F, -12.F) + pow(2.F, -13.F)
		+ pow(2.F, -16.F) + pow(2.F, -17.F)
		+ pow(2.F, -20.F) + pow(2.F, -21.F)
		+ pow(2.F, -23.F);

	std::bitset<sizeof(uint32_t) * CHAR_BIT > one_point_one(_one_point_one.i);
	std::cout << '\n' << "'one_point_one': " << one_point_one << '\n' // '0000 0000 0000 0001'
		<< "'check': " << check << '\n';

	return 0;
}
